// Copyright (C) 2019 Google LLC
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//      http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

#ifndef ICING_RESULT_RETRIEVER_H_
#define ICING_RESULT_RETRIEVER_H_

#include <utility>
#include <vector>

#include "utils/base/statusor.h"
#include "icing/proto/search.pb.h"
#include "icing/proto/term.pb.h"
#include "icing/query/query-terms.h"
#include "icing/schema/schema-store.h"
#include "icing/schema/section.h"
#include "icing/scoring/scored-document-hit.h"
#include "icing/snippet-retriever.h"
#include "icing/store/document-id.h"
#include "icing/store/document-store.h"
#include "icing/tokenization/language-segmenter.h"

namespace icing {
namespace lib {

class ResultRetriever {
 public:
  // Does not take any ownership, and all pointers must refer to valid objects
  // that outlive the one constructed.
  explicit ResultRetriever(const DocumentStore* doc_store,
                           const SchemaStore* schema_store,
                           const LanguageSegmenter* language_segmenter,
                           bool ignore_bad_document_ids = true)
      : doc_store_(*doc_store),
        snippet_retriever_(schema_store, language_segmenter),
        ignore_bad_document_ids_(ignore_bad_document_ids) {}

  // Gets results (pairs of DocumentProtos and SnippetProtos) with the given
  // document ids from the given document store. The
  // order of documents returned is the same as the order of document_ids.
  //
  // Parameter "ignore_bad_document_ids" indicates whether to ignore invalid and
  // non-existing document_ids. If it's true, errors on some document_ids will
  // be ignored and valid documents will be returned, otherwise any error will
  // be returned immediately. Note that IO errors will always be returned.
  //
  // Returns when ignore_bad_document_ids is true:
  //   A list of valid documents on success
  //   INTERNAL_ERROR on IO error
  //
  // Returns when ignore_bad_document_ids is false:
  //   A list of documents on success
  //   INVALID_ARGUMENT if any document_id < 0
  //   NOT_FOUND if any doc doesn't exist or has been deleted
  //   INTERNAL_ERROR on IO error
  libtextclassifier3::StatusOr<std::vector<SearchResultProto::ResultProto>>
  RetrieveResults(
      const ResultSpecProto& result_spec,
      const SectionRestrictQueryTermsMap& query_terms,
      TermMatchType::Code match_type,
      const std::vector<ScoredDocumentHit>& scored_document_hits) const;

 private:
  const DocumentStore& doc_store_;
  const SnippetRetriever snippet_retriever_;
  const bool ignore_bad_document_ids_;
};

}  // namespace lib
}  // namespace icing

#endif  // ICING_RESULT_RETRIEVER_H_
